# stdlib
import sys


__version__ = "0.2"

try:
    from .base import BaseAnonymizer, lazy_attribute  # noqa: F401
    from .utils import *  # noqa: F401,F403

except ImportError:
    # During setup.py not all dependencies may be installed, which may cause some
    # imports to fail. We still need to be able to import __init__ to check __version__,
    # as this is considered a good practice (having __version__ inside __init__)
    #
    # That's the cost of having shorthands like:
    #
    # >>> import anon
    # >>> anon.BaseAnonymizer
    #
    # versus having to import from base:
    #
    # >>> from anon.base import BaseAnonymizer
    # >>> BaseAnonymizer
    #
    if not sys.argv[0].endswith("setup.py"):
        raise



# stdlib
from collections import OrderedDict
from logging import getLogger

# deps
from chunkator import chunkator_page
from django_bulk_update.helper import bulk_update


logger = getLogger(__name__)


class OrderedDeclaration(object):
    """ Any classes inheriting from this will have an unique global counter
        associated with it. This counter is used to determine the order in
        which fields were declarated

        Idea taken from: https://stackoverflow.com/a/4460034/639465
        Also inspired by https://github.com/FactoryBoy/factory_boy
    """

    global_counter = 0

    def __init__(self):
        self._order = self.__class__.global_counter
        self.__class__.global_counter += 1


class LazyAttribute(OrderedDeclaration):
    def __init__(self, lazy_fn):
        super(LazyAttribute, self).__init__()
        self.lazy_fn = lazy_fn

    def __call__(self, *args, **kwargs):
        return self.lazy_fn(*args, **kwargs)


def lazy_attribute(lazy_fn):
    """ Returns LazyAttribute objects, that basically marks functions that
        should take `obj` as first parameter. This is useful when you need
        to take in consideration other values of `obj`

        Example:

        >>> full_name = lazy_attribute(o: o.first_name + o.last_name)

    """
    return LazyAttribute(lazy_fn)


class BaseAnonymizer(object):
    def run(self, select_chunk_size=None, **bulk_update_kwargs):
        self._declarations = self.get_declarations()

        queryset = self.get_queryset()
        update_fields = list(self._declarations.keys())
        update_batch_size = bulk_update_kwargs.pop(
            "batch_size", self._meta.update_batch_size
        )

        if select_chunk_size is None:
            select_chunk_size = self._meta.select_chunk_size

        if update_batch_size > select_chunk_size:
            raise ValueError(
                "update_batch_size ({}) should not be higher than "
                "select_chunk_size ({})".format(update_batch_size, select_chunk_size)
            )

        # info used in log messages
        model_name = self._meta.model.__name__
        current_batch = 0

        for page in chunkator_page(queryset, chunk_size=select_chunk_size):
            logger.info(
                "Updating {}... {}-{}".format(
                    model_name, current_batch, current_batch + select_chunk_size
                )
            )
            current_batch += select_chunk_size

            objs = []
            for obj in page:
                self.patch_object(obj)
                objs.append(obj)

            bulk_update(
                objs,
                **dict(
                    update_fields=update_fields,
                    batch_size=update_batch_size,
                    **bulk_update_kwargs
                )
            )

        if current_batch == 0:
            logger.info("{} has no records".format(model_name))

    def get_meta(self):
        meta = self.Meta()
        if not hasattr(meta, "select_chunk_size"):
            # Chunk size to iterate over
            meta.select_chunk_size = 5000
        if not hasattr(meta, "update_batch_size"):
            # Batch size for bulk updates
            meta.update_batch_size = 200
        return meta

    _meta = property(get_meta)

    def get_manager(self):
        meta = self._meta
        return getattr(meta, "manager", meta.model.objects)

    _manager = property(get_manager)

    def get_queryset(self):
        """ Override this if you want to delimit the objects that should be
            affected by anonymization
        """
        return self.get_manager().all()

    def patch_object(self, obj):
        """ Update object attributes with fake data provided by replacers
        """
        # using obj.__dict__ instead of getattr for performance reasons
        # see https://stackoverflow.com/a/9791053/639465
        fields = [field for field in self._declarations if obj.__dict__[field]]

        for field in fields:
            replacer = self._declarations[field]
            if isinstance(replacer, LazyAttribute):
                # Pass in obj for LazyAttributes
                new_value = replacer(obj)
            elif callable(replacer):
                new_value = replacer()
            else:
                new_value = replacer

            obj.__dict__[field] = new_value

        self.clean(obj)

    def clean(self, obj):
        """ Use this function if you need to update additional data that may
            rely on multiple fields, or if you need to update multiple fields
            at once
        """
        pass

    def get_declarations(self):
        """ Returns ordered declarations. Any non-ordered declarations, for
            example any types that does not inherit from OrderedDeclaration
            will come first, as they are considered "raw" values and should
            not be affected by the order of other non-ordered declarations
        """

        def _sort_declaration(declaration):
            name, value = declaration
            if isinstance(value, OrderedDeclaration):
                return value._order
            else:
                # Any non-ordered declarations come first
                return -1

        declarations = self._get_class_attributes().items()
        sorted_declarations = sorted(declarations, key=_sort_declaration)

        return OrderedDict(sorted_declarations)

    def _get_class_attributes(self):
        """ Return list of class attributes, which also includes methods and
            subclasses, ignoring any magic methods and reserved attributes
        """
        reserved_names = list(BaseAnonymizer.__dict__.keys()) + ["Meta"]

        return {
            name: self.__class__.__dict__[name]
            for name, value in self.__class__.__dict__.items()
            if not name.startswith("__") and name not in reserved_names
        }



# stdlib
import itertools
import random


_WORD_LIST = [
    "a",
    "ab",
    "accusamus",
    "accusantium",
    "ad",
    "adipisci",
    "alias",
    "aliquam",
    "aliquid",
    "amet",
    "animi",
    "aperiam",
    "architecto",
    "asperiores",
    "aspernatur",
    "assumenda",
    "at",
    "atque",
    "aut",
    "autem",
    "beatae",
    "blanditiis",
    "commodi",
    "consectetur",
    "consequatur",
    "consequuntur",
    "corporis",
    "corrupti",
    "culpa",
    "cum",
    "cumque",
    "cupiditate",
    "debitis",
    "delectus",
    "deleniti",
    "deserunt",
    "dicta",
    "dignissimos",
    "distinctio",
    "dolor",
    "dolore",
    "dolorem",
    "doloremque",
    "dolores",
    "doloribus",
    "dolorum",
    "ducimus",
    "ea",
    "eaque",
    "earum",
    "eius",
    "eligendi",
    "enim",
    "eos",
    "error",
    "esse",
    "est",
    "et",
    "eum",
    "eveniet",
    "ex",
    "excepturi",
    "exercitationem",
    "expedita",
    "explicabo",
    "facere",
    "facilis",
    "fuga",
    "fugiat",
    "fugit",
    "harum",
    "hic",
    "id",
    "illo",
    "illum",
    "impedit",
    "in",
    "incidunt",
    "inventore",
    "ipsa",
    "ipsam",
    "ipsum",
    "iste",
    "itaque",
    "iure",
    "iusto",
    "labore",
    "laboriosam",
    "laborum",
    "laudantium",
    "libero",
    "magnam",
    "magni",
    "maiores",
    "maxime",
    "minima",
    "minus",
    "modi",
    "molestiae",
    "molestias",
    "mollitia",
    "nam",
    "natus",
    "necessitatibus",
    "nemo",
    "neque",
    "nesciunt",
    "nihil",
    "nisi",
    "nobis",
    "non",
    "nostrum",
    "nulla",
    "numquam",
    "occaecati",
    "odio",
    "odit",
    "officia",
    "officiis",
    "omnis",
    "optio",
    "pariatur",
    "perferendis",
    "perspiciatis",
    "placeat",
    "porro",
    "possimus",
    "praesentium",
    "provident",
    "quae",
    "quaerat",
    "quam",
    "quas",
    "quasi",
    "qui",
    "quia",
    "quibusdam",
    "quidem",
    "quis",
    "quisquam",
    "quo",
    "quod",
    "quos",
    "ratione",
    "recusandae",
    "reiciendis",
    "rem",
    "repellat",
    "repellendus",
    "reprehenderit",
    "repudiandae",
    "rerum",
    "saepe",
    "sapiente",
    "sed",
    "sequi",
    "similique",
    "sint",
    "sit",
    "soluta",
    "sunt",
    "suscipit",
    "tempora",
    "tempore",
    "temporibus",
    "tenetur",
    "totam",
    "ullam",
    "unde",
    "ut",
    "vel",
    "velit",
    "veniam",
    "veritatis",
    "vero",
    "vitae",
    "voluptas",
    "voluptate",
    "voluptatem",
    "voluptates",
    "voluptatibus",
    "voluptatum",
]


try:
    xrange
except NameError:
    # Python 2/3 proof
    xrange = range


def _cycle_over_sample_range(start, end, sample_size):
    """
    Given a range (start, end), returns a generator that will cycle over a population
    sample with size specified by ``sample_size``
    """
    return itertools.cycle(random.sample(xrange(start, end), sample_size))


# Holds the average size of word sample
_avg_word_size = sum(map(len, _WORD_LIST)) / len(_WORD_LIST)

# Holds a generator that each iteration returns a different word
_word_generator = itertools.cycle(_WORD_LIST)

# Holds the size of smallest word in _WORD_LIST and is used to define bounds
_min_word_size = len(sorted(_WORD_LIST, key=lambda w: len(w))[0])

# Holds a generator that each iteration returns a different number
_number_generator = itertools.cycle("86306894249026785203141")

# Holds a generator for small integers, same as Django's PositiveSmallIntegerField
_small_int_generator = _cycle_over_sample_range(0, 32767, 1000)

# Holds a generator for small signed integers, same as Django's SmallIntegerField
_small_signed_int_generator = _cycle_over_sample_range(-32768, 32767, 1000)

# Holds a generator for integers, same as Django's PositiveIntegerField
_int_generator = _cycle_over_sample_range(0, 2147483647, 10000)

# Holds a generator for signed integers, same as Django's IntegerField
_signed_int_generator = _cycle_over_sample_range(-2147483648, 2147483647, 100000)


def fake_word(min_size=_min_word_size, max_size=20):
    """ Return fake word

    :min_size: Minimum number of chars
    :max_size: Maximum number of chars

    Example:

    >>> import django_anon as anon
    >>> print(anon.fake_word())
    adipisci

    """
    if min_size < _min_word_size:
        raise ValueError("no such word with this size < min_size")

    for word in _word_generator:
        if min_size <= len(word) <= max_size:
            return word


def fake_text(max_size=255, max_diff_allowed=5, separator=" "):
    """ Return fake text

    :max_size: Maximum number of chars
    :max_diff_allowed: Maximum difference (fidelity) allowed, in chars number
    :separator: Word separator

    Example:

    >>> print(anon.fake_text())
    alias aliquam aliquid amet animi aperiam architecto asperiores aspernatur assumenda at atque aut autem beatae blanditiis commodi consectetur consequatur consequuntur corporis corrupti culpa cum cumque cupiditate debitis delectus deleniti deserunt dicta

    """
    if max_diff_allowed < 1:
        raise ValueError("max_diff_allowed must be > 0")

    num_words = int(max_size / _avg_word_size)
    words = itertools.islice(_word_generator, num_words)

    text = separator.join(words)
    while len(text) > max_size:
        text = text[: text.rindex(separator)]

    return text


def fake_small_text(max_size=50):
    """ Preset for fake_text.

    :max_size: Maximum number of chars

    Example:

    >>> print(anon.fake_small_text())
    Distinctio Dolor Dolore Dolorem Doloremque Dolores

    """
    return fake_text(max_size=max_size).title()


def fake_name(max_size=15):
    """ Preset for fake_text. Also returns capitalized words.

    :max_size: Maximum number of chars

    Example:

    >>> print(anon.fake_name())
    Doloribus Ea

    """
    return fake_text(max_size=max_size).title()


def fake_username(max_size=10, separator=""):
    """ Returns fake username

    :max_size: Maximum number of chars
    :separator: Word separator
    :rand_range: Range to use when generating random number

    Example:

    >>> print(anon.fake_username())
    eius54455

    """
    random_number = str(next(_small_int_generator))
    min_size_allowed = _min_word_size + len(random_number)

    if max_size < min_size_allowed:
        raise ValueError("username must be >= {}".format(min_size_allowed))
    else:
        max_size -= len(random_number)

    return fake_text(max_size, separator=separator) + random_number


def fake_email(max_size=25, suffix="@example.com"):
    """ Returns fake email address

    :max_size: Maximum number of chars
    :suffix: Suffix to add to email addresses (including @)

    Example:

    >>> print(anon.fake_email())
    enim120238@example.com

    """
    min_size_allowed = _min_word_size + len(suffix)

    if max_size + len(suffix) > 254:
        # an email address must not exceed 254 chars
        raise ValueError("email address must not exceed 254 chars")
    elif max_size < min_size_allowed:
        raise ValueError("max_size must be >= {}".format(min_size_allowed))
    else:
        max_size -= len(suffix)

    return fake_username(max_size, separator=".") + suffix


def fake_url(max_size=50, scheme="http://", suffix=".com"):
    """ Returns fake URL

    :max_size: Maximum number of chars
    :scheme: URL scheme (http://)
    :suffix: Suffix to add to domain (including dot)

    Example:

    >>> print(anon.fake_url())
    http://facilis.fuga.fugiat.fugit.harum.hic.id.com

    """
    min_size_allowed = _min_word_size + len(scheme) + len(suffix)

    if max_size < min_size_allowed:
        raise ValueError("max_size must be >= {}".format(min_size_allowed))
    else:
        max_size -= len(scheme) + len(suffix)

    domain = fake_text(max_size=max_size, separator=".") + suffix
    return scheme + domain


def fake_phone_number(format="999-999-9999"):
    """ Returns a fake phone number in the desired format

    :format: Format of phone number to generate

    Example:

    >>> print(anon.fake_phone_number())
    863-068-9424

    """
    number = []
    for char in format:
        if char == "9":
            n = next(_number_generator)
            if not number:
                # do not start phone numbers with zero
                while n == "0":
                    n = next(_number_generator)
            number.append(n)
        else:
            number.append(char)
    return "".join(number)



# stdlib
import re

# deps
from django.test import TestCase

# local
from anon import utils


class UtilsTestCase(TestCase):
    def test_fake_word(self):
        text = utils.fake_word(min_size=6)
        self.assertGreaterEqual(len(text), 6)

    def test_fake_text(self):
        text = utils.fake_text(40)
        self.assertLessEqual(len(text), 40)

    def test_fake_text_separator(self):
        text = utils.fake_text(40, separator="...")
        self.assertIn("...", text)

    def test_fake_name(self):
        text = utils.fake_name(15)
        self.assertLessEqual(len(text), 15)

    def test_fake_username(self):
        text = utils.fake_username(45, separator="_")
        self.assertIn("_", text)
        self.assertLessEqual(len(text), 45)

    def test_fake_email(self):
        text = utils.fake_email(20)
        self.assertLessEqual(len(text), 20)

    def test_fake_url(self):
        text = utils.fake_url(30, scheme="https://", suffix=".com.br")
        self.assertLessEqual(len(text), 30)
        self.assertTrue(text.startswith("https://"))
        self.assertIn(".com.br", text)

    def test_fake_phone_number(self):
        text = utils.fake_phone_number(format="(99) 9999-9999")
        self.assertTrue(bool(re.match(r"^\(\d{2}\) \d{4}-\d{4}$", text)))



# deps
from django.test import TestCase

# local
import anon

from .compat import mock


class BaseAnonymizer(anon.BaseAnonymizer):
    class Meta:
        pass

    def __init__(self):
        """
        The reason we set declarations manually in ``__init__`` is because we want to
        patch objects individually for unit testing purposes, without using the
        ``run()`` method. In general cases, declarations will be set by ``run()``
        """
        self._declarations = self.get_declarations()


class BaseTestCase(TestCase):
    def test_get_meta(self):
        class Anon(BaseAnonymizer):
            class Meta:
                pass

        anonymizer = Anon()
        self.assertIsInstance(anonymizer._meta, Anon.Meta)

    def test_get_manager(self):
        manager = object()

        m = mock.Mock()
        m.other_manager = manager

        class Anon(BaseAnonymizer):
            class Meta:
                model = m
                manager = m.other_manager

        anonymizer = Anon()
        self.assertEqual(anonymizer.get_manager(), manager)

    def test_get_queryset(self):
        sample_obj = object()

        m = mock.Mock()
        m.objects.all.return_value = [sample_obj]

        class Anon(BaseAnonymizer):
            class Meta:
                model = m

        anonymizer = Anon()
        result = anonymizer.get_queryset()
        self.assertEqual(result, [sample_obj])
        m.objects.all.assert_called_once()

    def test_patch_object(self):
        fake_first_name = lambda: "foo"  # noqa: E731

        class Anon(BaseAnonymizer):
            first_name = fake_first_name
            last_name = fake_first_name
            raw_data = {1: 2}

        class Obj(object):
            def __init__(self):
                self.first_name = "zzz"
                self.last_name = ""  # empty data should be kept empty
                self.raw_data = {}

        obj = Obj()

        anonymizer = Anon()
        anonymizer.patch_object(obj)
        self.assertEqual(obj.first_name, "foo")
        self.assertEqual(obj.last_name, "")
        self.assertEqual(obj.raw_data, {})

    @mock.patch("anon.base.bulk_update")
    @mock.patch("anon.base.chunkator_page")
    def test_run(self, chunkator_page, bulk_update):
        class Anon(BaseAnonymizer):
            class Meta:
                model = mock.Mock(__name__="x")
                update_batch_size = 42

            first_name = "xyz"

        class Obj(object):
            def __init__(self):
                self.first_name = "zzz"

        obj = Obj()

        chunkator_page.return_value = [[obj]]

        anonymizer = Anon()
        anonymizer.get_queryset = mock.Mock(return_value=[obj])
        anonymizer.patch_object = mock.Mock()
        anonymizer.run()

        anonymizer.patch_object.assert_called_once_with(obj)
        bulk_update.assert_called_once_with(
            [obj], batch_size=42, update_fields=["first_name"]
        )

    def test_lazy_attribute(self):
        lazy_fn = mock.Mock()
        fake_first_name = anon.lazy_attribute(lazy_fn)

        class Anon(BaseAnonymizer):
            first_name = fake_first_name

        class Obj(object):
            def __init__(self):
                self.first_name = "zzz"

        obj = Obj()

        anonymizer = Anon()
        anonymizer.patch_object(obj)
        lazy_fn.assert_called_once_with(obj)

    def test_lazy_attribute_decorator(self):
        lazy_fn = mock.Mock()

        class Anon(BaseAnonymizer):
            @anon.lazy_attribute
            def first_name(self):
                return lazy_fn(self)

        class Obj(object):
            def __init__(self):
                self.first_name = "zzz"

        obj = Obj()

        anonymizer = Anon()
        anonymizer.patch_object(obj)
        lazy_fn.assert_called_once_with(obj)

    def test_raw_attributes(self):
        class Anon(BaseAnonymizer):
            raw_data = {}

        class Obj(object):
            def __init__(self):
                self.raw_data = {"password": "xyz"}

        obj = Obj()

        anonymizer = Anon()
        anonymizer.patch_object(obj)
        self.assertEqual(obj.raw_data, {})

    def test_clean(self):
        class Anon(BaseAnonymizer):
            line1 = ""
            line2 = ""
            line3 = ""

            def clean(self, obj):
                obj.line1 = "foo"
                obj.line2 = "bar"

        class Obj(object):
            def __init__(self):
                self.line1 = "X"
                self.line2 = "Y"
                self.line3 = "Z"

        obj = Obj()

        anonymizer = Anon()
        anonymizer.patch_object(obj)
        self.assertEqual(obj.line1, "foo")
        self.assertEqual(obj.line2, "bar")
        self.assertEqual(obj.line3, "")

    def test_get_declarations(self):
        # Ensure the order is preserved
        class Anon(BaseAnonymizer):
            a = anon.lazy_attribute(lambda o: 4)
            c = anon.lazy_attribute(lambda o: 6)
            b = anon.lazy_attribute(lambda o: 5)

            def get_queryset(self):
                return []

        anonymizer = Anon()
        self.assertEqual(list(anonymizer.get_declarations().keys()), ["a", "c", "b"])



# stdlib
import os

# deps
from setuptools import Command, find_packages, setup


# Dynamically calculate the version based on anon.VERSION
VERSION = __import__("anon").__version__


with open("README.rst") as readme_file:

    def remove_banner(readme):
        # Since PyPI does not support raw directives, we remove them from the README
        #
        # raw directives are only used to make README fancier on GitHub and do not
        # contain relevant information to be displayed in PyPI, as they are not tied
        # to the current version, but to the current development status
        out = []
        lines = iter(readme.splitlines(True))
        for line in lines:
            if line.startswith(".. BANNERSTART"):
                for line in lines:
                    if line.strip() == ".. BANNEREND":
                        break
            else:
                out.append(line)
        return "".join(out)

    README = remove_banner(readme_file.read())


class PublishCommand(Command):
    description = "Publish to PyPI"
    user_options = []

    def run(self):
        # Build & Upload
        # https://packaging.python.org/tutorials/packaging-projects/#generating-distribution-archives
        # https://packaging.python.org/tutorials/packaging-projects/#uploading-the-distribution-archives
        os.system("python setup.py sdist bdist_wheel")
        os.system("twine upload dist/*")

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass


class CreateTagCommand(Command):
    description = "Create release tag"
    user_options = []

    def run(self):
        os.system("git tag -a %s -m 'v%s'" % (VERSION, VERSION))
        os.system("git push --tags")

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass


setup(
    name="django-anon",
    version=VERSION,
    packages=find_packages(),
    install_requires=["django-bulk-update", "django-chunkator<2"],
    cmdclass={"publish": PublishCommand, "tag": CreateTagCommand},
    # metadata for upload to PyPI
    description="Anonymize production data so it can be safely used in not-so-safe environments",
    long_description=README,
    long_description_content_type="text/x-rst",
    author="Tesorio",
    author_email="hello@tesorio.com",
    url="https://github.com/Tesorio/django-anon",
    license="MIT",
    platforms=["any"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Framework :: Django",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
    ],
)
